#Base Types
EXCEPTION = "exception"
ALLOWANCE = "allowance"

ASSIGNATION = "assignation"
PROVISION = "provision"
ASSIGNATION_REQUEST = "assignation_request"
PROVISION_REQUEST = "provision_request"


ACTIVATE_POD = "activate_pod"
DEACTIVATE_POD ="deactivate_pod"


# ASIGN PATH
ASSIGN = "assign"
CANCEL_ASSIGN = "cancel_assign"


# Arkitekt
BOUNCED_ASSIGN = "bounced_assign"
BOUNCED_CANCEL_ASSIGN = "bounced_cancel_assign"

# Pod 
ASSIGN_DONE = "assign_done"
ASSIGN_PROGRES ="assign_progress"
ASSIGN_CRITICAL = "assign_fail"
ASSIGN_YIELD =  "assign_yield"
ASSIGN_RETURN = "assign_return"
ASSIGN_RETRY = "assign_retry"

# Reserve Path
RESERVE = "reserve"
CANCEL_RESERVE = "cancel_reserve"

# Arkitekt
BOUNCED_RESERVE = "bounced_reserve"
BOUNCED_CANCEL_RESERVE = "bounced_cancel_reserve"

# Provider
RESERVE_DONE = "reserve_done"
RESERVE_PROGRESS = "reserve_progress"
RESERVE_FAIL = "reserve_fail"
RESERVE_CRITICAL = "reserve_critical"


UNRESERVE = "unreserve"

# Provide Path
PROVIDE = "provide"
CANCEL_PROVIDE = "cancel_provide"

# Arkitekt
BOUNCED_PROVIDE = "bounced_provide"
BOUNCED_CANCEL_PROVIDE = "bounced_cancel_provide"

# Provider
PROVIDE_DONE = "provide_done"
PROVIDE_PROGRESS = "provide_progress"
PROVIDE_FAIL = "provide_fail"
PROVIDE_CRITICAL = "provide_critical"



UNPROVIDE = "unprovide"